
#include "mod/mod.h"
#include "mod/mtypes.h"
#include "fxttypes.h"  // ulong

void
chebyU2(umod_t n, umod_t &u1, umod_t &u0, umod_t m)
// Compute  u1=U(n-1,x), u0=U(n,x) modulo m where x=2
// and U(n) is the n-th Chebyshev polynomial of the second kind.
// Must have m>4
{
    if ( n<=1 )
    {
        if ( 1==n )  { u1 = 1; u0 = 4; /* 4==2*x */ }
        else { u1 = 0; u0 = 1; }
        return;
    }

    ulong q = n & 1;
    umod_t nr = (n-q) >> 1;
    q ^= 1UL; // whether n is even
    umod_t ru1, ru0;
    chebyU2(nr, ru1, ru0, m);  // recursion
    umod_t s1 = mul_mod(ru1, ru1, m); // s1=ru1^2
    umod_t s2 = mul_mod(ru0, ru0, m); // s2=ru0^2
    umod_t s3 = add_mod(ru0, ru1, m);
    s3 = mul_mod(s3, s3, m); // s3=(ru0+ru1)^2
    umod_t ds = sub_mod(s2, s1, m);  // ds=s2-s1
    if ( q )
    {
        u0 = ds;
        umod_t t = sub_mod(s3, s2, m);
        umod_t p = s1;
        p = add_mod(p,p,m);
        p = add_mod(p,p,m);
        p = add_mod(p,s1,m); // == 5*s1 == (2*x+1)*s1
        u1 = sub_mod(t, p, m);  // u1 = s3-s2-((2*x)+1)*s1
    }
    else
    {
        u1 = ds;
        umod_t t = sub_mod(s1, s3, m);
        umod_t p = s2;
        p = add_mod(p,p,m);
        p = add_mod(p,p,m);
        p = add_mod(p,s2,m); // == 5*s2 == (2*x+1)*s2
        u0 = add_mod(t, p, m); // u0 = s1+((2*x)+1)*s2-s3
    }
    return;
}
// -------------------------


umod_t
chebyU2(umod_t n, umod_t m)
// Return u=U(n,x) modulo m where x=2
// and U(n) is the n-th Chebyshev polynomial of the second kind.
// Must have m>4
{
    umod_t u1, u0;
    chebyU2(n, u1, u0, m);
    return u0;
}
// -------------------------


void
chebyT2(umod_t n, umod_t &t1, umod_t &t0, umod_t m)
// Compute t1=T(n-1,x), t0=T(n,x) modulo m where x=2
// and T(n) is the n-th Chebyshev polynomial of the first kind.
// Must have m>4
// T_{n} = U_{n} - x*U_{n-1}  (where x=2)
// T_{n-1} = U_{n-1} - x*U_{n-2}
//  (where U_{n-2}=2*x*U_{n-1}-U_{n})
{
    umod_t u1, u0;
    chebyU2(n, u1, u0, m);
    umod_t r0 = u0;  r0=sub_mod(r0, u1, m);  r0=sub_mod(r0, u1, m);
    t0 = r0;
    umod_t u2 = u1;  u2 = add_mod(u2, u2, m);  u2 = add_mod(u2, u2, m);
    u2 = sub_mod(u2, u0, m);
    umod_t r1 = u1;  r1 = sub_mod(r1, u2, m);  r1 = sub_mod(r1, u2, m);
    t1 = r1;
    return;
}
// -------------------------


umod_t
chebyT2(umod_t n, umod_t m)
// Return t=T(n,x) modulo m where x=2
// and T(n) is the n-th Chebyshev polynomial of the first kind.
// Must have m>4
// T_{n} = U_{n} - x*U_{n-1}  (where x=2)
{
    umod_t u1, u0;
    chebyU2(n, u1, u0, m);
    umod_t r0 = u0;  r0=sub_mod(r0, u1, m);  r0=sub_mod(r0, u1, m);
    return r0;
}
// -------------------------
